/*
Module : 8055N++.h
Purpose: Defines the interface for a set of C++ class which encapsulates the encapsulates the K8055, K8055N, 
         WPI110N, WSI8055N, VM110 & VM110N USB ExperimentExperiment Interface boards from Velleman / Whadda 
         using the Windows HID APIs.
History: PJN / 30-03-2023 1. Initial release

Copyright (c) 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


/////////////////////////// Macros / Defines //////////////////////////////////

#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif //#if defined(_MSC_VER) && (_MSC_VER >= 1020)

#ifndef __8055NWRAPPERS_H__
#define __8055NWRAPPERS_H__

#ifndef C8055NWRAPPERS_EXT_CLASS
#define C8055NWRAPPERS_EXT_CLASS
#endif //#ifndef C8055NWRAPPERS_EXT_CLASS

#pragma comment(lib, "hid.lib")
#pragma comment(lib, "SetupAPI.lib")


/////////////////////////// Includes //////////////////////////////////////////

#include "HIDWrappers.h"

#ifndef _ARRAY_
#pragma message("To avoid this message, please put array in your pre compiled header (normally stdafx.h)")
#include <array>
#endif //#ifndef _ARRAY_
#include <cmath>


/////////////////////////// Classes ///////////////////////////////////////////

namespace HID
{

//Wrapper for a HID handle
class C8055NWRAPPERS_EXT_CLASS C8055NDevice : public HID::CDevice
{
public:
//Enums
  enum class CARD_VERSION
  {
    UNKNOWN = 0,
    K8055 = 1,
    K8055N = 2
  };

//Constructors / Destructors
  C8055NDevice() = default;
  C8055NDevice(_In_ const C8055NDevice& device) = delete;
  C8055NDevice(_In_ C8055NDevice&& device) = delete;

//Methods
  C8055NDevice& operator=(_In_ const C8055NDevice& device) = delete;
  C8055NDevice& operator=(_In_ C8055NDevice&& device) = delete;

  static _Success_(return != false) bool Enumerate(_Inout_ StringArray& deviceNames, _In_ DWORD dwFlags = DIGCF_PRESENT | DIGCF_DEVICEINTERFACE, _In_opt_ LPCTSTR pszEnumerator = nullptr, _In_ USHORT nVendorID = 0x10CF)
  {
    return EnumerateDevices(deviceNames, [nVendorID](_In_z_ const auto& szDevicePath) noexcept
      {
#pragma warning(suppress: 28193)
        const int nProductID{GetProductID(szDevicePath)};
        return (GetVendorID(szDevicePath) == nVendorID) && ((nProductID == 0x5500) || (nProductID == 0x5501) || (nProductID == 0x5502) || (nProductID == 0x5503));
      }, dwFlags, pszEnumerator);
  }

  _Success_(return != false) bool Open(_In_ LPCTSTR pszDeviceName, _In_ DWORD dwDesiredAccess = GENERIC_READ | GENERIC_WRITE, _In_ DWORD dwShareMode = FILE_SHARE_READ | FILE_SHARE_WRITE, _In_opt_ LPSECURITY_ATTRIBUTES lpSecurityAttributes = nullptr, _In_ DWORD dwFlagsAndAttributes = 0) noexcept
  {
    //Delegate to the base class
    if (!__super::Open(pszDeviceName, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwFlagsAndAttributes))
      return false;

    //Try switching the card to the new K8055N protocol
#pragma warning(suppress: 26446)
    m_TxBuffer[1] = 6;
    if (!OutIn(true, false))
    {
      Close();
      return false;
    }
    if (!OutIn(false, true))
    {
      Close();
      return false;
    }
#pragma warning(suppress: 26446)
    if (m_RxBuffer[2] > 10)
    {
#pragma warning(suppress: 26446)
      if (m_RxBuffer[2] <= 20) //Was in old mode
        OutIn(true, false);
#pragma warning(suppress: 26446)
      while (m_RxBuffer[2] <= 20)
      {
        OutIn(false, true);
      }
      m_bNewMode = true;
    }
    else
      m_bNewMode = false;

    //Determine the card address
#pragma warning(suppress: 28193)
    const int nProductID{GetProductID(pszDeviceName)};
    m_nCardAddress = nProductID - 0x5500;

    return true;
  }

  void Close() noexcept
  {
    //Delegate to the base class
    __super::Close();

    //Reset our member variables
    m_RxBuffer = {};
    m_TxBuffer = {};
    m_nCardAddress = -1;
    m_bNewMode = false;
  }

  _Success_(return != false) bool SetCounterDebounceTime(_In_ int nCounter, _In_ int nDebounceTime) noexcept
  {
    //Validate our parameters
    if ((nCounter > 2) || (nCounter < 1))
    {
      SetLastError(ERROR_INVALID_PARAMETER);
      return false;
    }

    int nT1{static_cast<int>(0.9 * 2.4541 * std::pow(nDebounceTime, 0.5328))};
    if (nT1 < 1)
      nT1 = 1;
    if (nT1 > 255)
      nT1 = 255;
    if (m_bNewMode)
#pragma warning(suppress: 26446 26472)
      m_TxBuffer[1] = static_cast<BYTE>(static_cast<int>(TAG::NEW_SET_DEBOUNCE_COUNTER_1) + (nCounter - 1));
    else
#pragma warning(suppress: 26446 26472)
      m_TxBuffer[1] = static_cast<BYTE>(static_cast<int>(TAG::OLD_SET_DEBOUNCE_COUNTER_1) + (nCounter - 1));
#pragma warning(suppress: 26446 26472 26482)
    m_TxBuffer[static_cast<size_t>(nCounter) + 6] = static_cast<BYTE>(nT1);
    return OutIn(true, false);
  }

  _Success_(return != false) bool ResetCounter(_In_ int nCounter) noexcept
  {
    //Validate our parameters
    if ((nCounter > 2) || (nCounter < 1))
    {
      SetLastError(ERROR_INVALID_PARAMETER);
      return false;
    }

    if (m_bNewMode)
#pragma warning(suppress: 26446 26472)
      m_TxBuffer[1] = static_cast<BYTE>(static_cast<int>(TAG::NEW_RESET_COUNTER_1) + (nCounter - 1));
    else
#pragma warning(suppress: 26446 26472)
      m_TxBuffer[1] = static_cast<BYTE>(static_cast<int>(TAG::OLD_RESET_COUNTER_1) + (nCounter - 1));
    return OutIn(true, false);
  }

  _Success_(return != false) bool OutputAnalogChannel(_In_ int nChannel, _In_ BYTE Data) noexcept
  {
    //Validate our parameters
    if ((nChannel > 2) || (nChannel < 1))
    {
      SetLastError(ERROR_INVALID_PARAMETER);
      return false;
    }

#pragma warning(suppress: 26446 26482)
    m_TxBuffer[static_cast<size_t>(nChannel) + 2] = Data;
    if (m_bNewMode)
#pragma warning(suppress: 26446)
      m_TxBuffer[1] = static_cast<BYTE>(TAG::NEW_SEND_OTHER_DATA);
    else
#pragma warning(suppress: 26446)
      m_TxBuffer[1] = static_cast<BYTE>(TAG::OLD_SEND_OTHER_DATA);
    return OutIn(true, false);
  }

  _Success_(return != false) bool OutputAllAnalog(_In_ BYTE Data1, _In_ BYTE Data2) noexcept
  {
#pragma warning(suppress: 26446)
    m_TxBuffer[3] = Data1;
#pragma warning(suppress: 26446)
    m_TxBuffer[4] = Data2;
    if (m_bNewMode)
#pragma warning(suppress: 26446)
      m_TxBuffer[1] = static_cast<BYTE>(TAG::NEW_SEND_OTHER_DATA);
    else
#pragma warning(suppress: 26446)
      m_TxBuffer[1] = static_cast<BYTE>(TAG::OLD_SEND_OTHER_DATA);
    return OutIn(true, false);
  }

  _Success_(return != false) bool ClearAnalogChannel(_In_ int nChannel) noexcept
  {
    return OutputAnalogChannel(nChannel, 0);
  }

  _Success_(return != false) bool ClearAllAnalog() noexcept
  {
    return OutputAllAnalog(0, 0);
  }

  _Success_(return != false) bool SetAnalogChannel(_In_ int nChannel) noexcept
  {
    return OutputAnalogChannel(nChannel, 255);
  }

  _Success_(return != false) bool SetAllAnalog() noexcept
  {
    return OutputAllAnalog(255, 255);
  }

  _Success_(return != false) bool WriteAllDigital(_In_ BYTE Data) noexcept
  {
#pragma warning(suppress: 26446)
    m_TxBuffer[2] = Data;
    m_LastDigitalOutValue = Data;

    if (m_bNewMode)
#pragma warning(suppress: 26446)
      m_TxBuffer[1] = static_cast<BYTE>(TAG::NEW_SEND_OTHER_DATA);
    else
#pragma warning(suppress: 26446)
      m_TxBuffer[1] = static_cast<BYTE>(TAG::OLD_SEND_OTHER_DATA);
    return OutIn(true, false);
  }

  _Success_(return != false) bool ClearDigitalChannel(_In_ int nChannel) noexcept
  {
    //Validate our parameters
    if ((nChannel > 8) || (nChannel < 1))
    {
      SetLastError(ERROR_INVALID_PARAMETER);
      return false;
    }

#pragma warning(suppress: 26472)
    const BYTE k{static_cast<BYTE>(1 << (nChannel - 1))};
#pragma warning(suppress: 26446)
    return WriteAllDigital(m_TxBuffer[2] & ~k);
  }

  _Success_(return != false) bool ClearAllDigital() noexcept
  {
    return WriteAllDigital(0);
  }

  _Success_(return != false) bool SetDigitalChannel(_In_ int nChannel) noexcept
  {
    //Validate our parameters
    if ((nChannel > 8) || (nChannel < 1))
    {
      SetLastError(ERROR_INVALID_PARAMETER);
      return false;
    }

#pragma warning(suppress: 26472)
    const BYTE k{static_cast<BYTE>(1 << (nChannel - 1))};
#pragma warning(suppress: 26446)
    return WriteAllDigital(m_TxBuffer[2] |= k);
  }

  _Success_(return != false) bool SetAllDigital() noexcept
  {
    return WriteAllDigital(255);
  }

  _Success_(return != false) bool ReadDigitalChannel(_In_ int nChannel, _Out_ bool& bValue) noexcept
  {
    //Validate our parameters
    if ((nChannel > 5) || (nChannel < 1))
    {
      SetLastError(ERROR_INVALID_PARAMETER);
      return false;
    }

    int nValue{0};
    if (!ReadAllDigital(nValue))
      return false;
    bValue = (nValue & (1 << (nChannel - 1))) != 0;
    return true;
  }

  _Success_(return != false) bool ReadAllDigital(_Out_ int& nValue) noexcept
  {
    if (m_bNewMode)
    {
#pragma warning(suppress: 26446)
      m_TxBuffer[1] = static_cast<BYTE>(TAG::NEW_READ_DIGITAL_INPUT);
      if (!OutIn(true, true))
        return false;
    }
    else
    {
      if (!OutIn(false, true))
        return false;
    }
#pragma warning(suppress: 26446)
    const int K1{m_RxBuffer[1]};
    nValue = (((K1 & 0x10) >> 4) |
      ((K1 & 0x20) >> 4) |
      ((K1 & 0x01) << 2) |
      ((K1 & 0x40) >> 3) |
      ((K1 & 0x80) >> 3));
    return true;
  }

  _Success_(return != false) bool ReadAnalogChannel(_In_ int nChannel, _Out_ BYTE& value) noexcept
  {
    //Validate our parameters
    if ((nChannel > 2) || (nChannel < 1))
    {
      SetLastError(ERROR_INVALID_PARAMETER);
      return false;
    }

    if (m_bNewMode)
    {
#pragma warning(suppress: 26446)
      m_TxBuffer[1] = static_cast<BYTE>(TAG::NEW_READ_ALL_ANALOG);
      if (!OutIn(true, true))
        return false;
    }
    else
    {
      if (!OutIn(false, true))
        return false;
    }
#pragma warning(suppress: 26446 26482)
    value = m_RxBuffer[static_cast<size_t>(nChannel) + 2];
    return true;
  }

  _Success_(return != false) bool ReadAllAnalog(_Out_ BYTE& Data1, _Out_ BYTE& Data2) noexcept
  {
    if (m_bNewMode)
    {
#pragma warning(suppress: 26446)
      m_TxBuffer[1] = static_cast<BYTE>(TAG::NEW_READ_ALL_ANALOG);
      if (!OutIn(true, true))
        return false;
    }
    else
    {
      if (!OutIn(false, true))
        return false;
    }
#pragma warning(suppress: 26446)
    Data1 = m_RxBuffer[3];
#pragma warning(suppress: 26446)
    Data2 = m_RxBuffer[4];
    return true;
  }

  _Success_(return != false) bool ReadCounter(_In_ int nCounter, _Out_ DWORD& nValue) noexcept
  {
    //Validate our parameters
    if ((nCounter > 2) || (nCounter < 1))
    {
      SetLastError(ERROR_INVALID_PARAMETER);
      return false;
    }

    if (m_bNewMode)
    {
#pragma warning(suppress: 26446 26472)
      m_TxBuffer[1] = static_cast<BYTE>(static_cast<int>(TAG::NEW_READ_32BIT_COUNTER_1) + (nCounter - 1));
      if (!OutIn(true, true))
        return false;
#pragma warning(suppress: 26446)
      nValue = m_RxBuffer[5] + (static_cast<DWORD>(m_RxBuffer[6]) << 8) + (static_cast<DWORD>(m_RxBuffer[7]) << 16) + (static_cast<DWORD>(m_RxBuffer[8]) << 24);
    }
    else
    {
      if (!OutIn(false, true))
        return false;
#pragma warning(suppress: 26446 26482)
      nValue = m_RxBuffer[(static_cast<size_t>(nCounter)*2) + 3] + (static_cast<DWORD>(m_RxBuffer[(static_cast<size_t>(nCounter)*2) + 4]) << 8);
    }
    return true;
  }

  _Success_(return != false) bool ReadCounter16(_In_ int nCounter, _Out_ USHORT& nValue) noexcept
  {
    if (m_bNewMode)
    {
#pragma warning(suppress: 26446 26472)
      m_TxBuffer[1] = static_cast<BYTE>(static_cast<int>(TAG::NEW_READ_16BIT_COUNTER_1) + (nCounter - 1));
      if (!OutIn(true, true))
        return false;
    }
    else
    {
      if (!OutIn(false, true))
        return false;
    }
#pragma warning(suppress: 26446 26482)
    nValue = m_RxBuffer[(static_cast<size_t>(nCounter) * 2) + 3] + (static_cast<DWORD>(m_RxBuffer[(static_cast<size_t>(nCounter) * 2) + 4]) << 8);
    return true;
  }

  _Success_(return != false) bool ReadAll(_Out_writes_bytes_(8) BYTE* pData) noexcept
  {
    //Validate our parameters
    if (pData == nullptr)
    {
      SetLastError(ERROR_INVALID_PARAMETER);
      return false;
    }

    if (m_bNewMode)
    {
#pragma warning(suppress: 26446)
      m_TxBuffer[1] = static_cast<BYTE>(TAG::NEW_READ_ALL);
      if (!OutIn(true, true))
        return false;
    }
    else
    {
      if (!OutIn(false, true))
        return false;
    }
#pragma warning(suppress: 26481)
    memcpy_s(pData, 8, m_RxBuffer.data() + 1, m_RxBuffer.size() - 1);
    return true;
  }

  _Success_(return != false) bool SetPWM(_In_ int nChannel, _In_ BYTE Data, _In_ BYTE Frequency) noexcept
  {
    //Validate our parameters
    if ((nChannel > 2) || (nChannel < 1))
    {
      SetLastError(ERROR_INVALID_PARAMETER);
      return false;
    }
    if ((Frequency > 3) || (Frequency < 1))
    {
      SetLastError(ERROR_INVALID_PARAMETER);
      return false;
    }
    if (!m_bNewMode)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return false;
    }

    std::array<BYTE, 5> tmpTxBuffer{};
    for (int i=2; i<=4; i++)
#pragma warning(suppress: 26446 26482)
      tmpTxBuffer[i] = m_TxBuffer[i];
#pragma warning(suppress: 26446)
    m_TxBuffer[1] = static_cast<BYTE>(TAG::NEW_SET_PWN);
#pragma warning(suppress: 26446 26472)
    m_TxBuffer[2] = static_cast<BYTE>(nChannel - 1);
#pragma warning(suppress: 26446)
    m_TxBuffer[3] = Data;
#pragma warning(suppress: 26446)
    m_TxBuffer[4] = Frequency;
    const bool bSuccess = OutIn(true, false);
    for (int i=2; i<=4; i++)
#pragma warning(suppress: 26446 26482)
      m_TxBuffer[i] = tmpTxBuffer[i];
    return bSuccess;
  }

  _Success_(return != false) bool ReadBackDigitalOut(_Out_ int& nValue) noexcept
  {
    if (m_bNewMode)
    {
#pragma warning(suppress: 26446)
      m_TxBuffer[1] = static_cast<BYTE>(TAG::NEW_READ_DIGITAL_OUT);
      if (!OutIn(true, true))
        return false;
#pragma warning(suppress: 26446)
      nValue = m_RxBuffer[5];
    }
    else
      nValue = m_LastDigitalOutValue;
    return true;
  }

  _Success_(return != false) bool ReadBackAnalogOut(_Out_ int& nValue1, _Out_ int& nValue2) noexcept
  {
    if (m_bNewMode)
    {
#pragma warning(suppress: 26446)
      m_TxBuffer[1] = static_cast<BYTE>(TAG::NEW_READ_ANALOG_OUT);
      if (!OutIn(true, true))
        return false;
#pragma warning(suppress: 26446)
      nValue1 = m_RxBuffer[6];
#pragma warning(suppress: 26446)
      nValue2 = m_RxBuffer[7];
    }
    else
    {
#pragma warning(suppress: 26446)
      nValue1 = m_TxBuffer[3];
#pragma warning(suppress: 26446)
      nValue2 = m_TxBuffer[4];
    }
    return true;
  }

  _Success_(return != false) bool Connected() noexcept
  {
    CARD_VERSION cardVersion{CARD_VERSION::UNKNOWN};
    if (!CardVersion(cardVersion))
      return false;
    return ((cardVersion == CARD_VERSION::K8055) || (cardVersion == CARD_VERSION::K8055N));
  }

  _Success_(return != false) bool CardVersion(_Out_ CARD_VERSION& cardVersion) noexcept
  {
    if (m_bNewMode)
    {
#pragma warning(suppress: 26446)
      m_TxBuffer[1] = static_cast<BYTE>(TAG::NEW_READ_DIGITAL_INPUT);
      if (!OutIn(true, true))
        return false;
#pragma warning(suppress: 26446)
      if (m_RxBuffer[2] == (m_nCardAddress + 21))
        cardVersion = CARD_VERSION::K8055N;
      else
        cardVersion = CARD_VERSION::UNKNOWN;
    }
    else
    {
      if (!OutIn(false, true))
        return false;
#pragma warning(suppress: 26446)
      if (m_RxBuffer[2] == (m_nCardAddress + 1))
        cardVersion = CARD_VERSION::K8055;
      else
        cardVersion = CARD_VERSION::UNKNOWN;
    }
    return true;
  }

protected:
//Enums
  enum class TAG
  {
    OLD_SET_DEBOUNCE_COUNTER_1 = 1,
    OLD_SET_DEBOUNCE_COUNTER_2 = 2,
    OLD_RESET_COUNTER_1 = 3,
    OLD_RESET_COUNTER_2 = 4,
    OLD_SEND_OTHER_DATA = 5,
    NEW_SET_DEBOUNCE_COUNTER_1 = 11,
    NEW_SET_DEBOUNCE_COUNTER_2 = 12,
    NEW_RESET_COUNTER_1 = 13,
    NEW_RESET_COUNTER_2 = 14,
    NEW_SEND_OTHER_DATA = 15,
    NEW_READ_DIGITAL_INPUT = 16,
    NEW_READ_ALL_ANALOG = 17,
    NEW_READ_32BIT_COUNTER_1 = 18,
    NEW_READ_32BIT_COUNTER_2 = 19,
    NEW_READ_16BIT_COUNTER_1 = 20,
    NEW_READ_16BIT_COUNTER_2 = 21,
    NEW_READ_ALL = 22,
    NEW_SET_PWN = 23,
    NEW_READ_DIGITAL_OUT = 24,
    NEW_READ_ANALOG_OUT = 25
  };

//Methods
  _Success_(return != false) bool WriteOutputReport(_In_reads_bytes_opt_(dwNumberOfBytesToWrite) const BYTE* byReport, _In_ DWORD dwNumberOfBytesToWrite) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_h != INVALID_HANDLE_VALUE);

    DWORD dwWritten = 0;
#pragma warning(suppress: 26472)
    return WriteFile(*this, byReport, dwNumberOfBytesToWrite, &dwWritten, nullptr) != FALSE;
  }

  _Success_(return != false) bool ReadInputReport(_Out_writes_bytes_to_opt_(nNumberOfBytesToRead, *lpNumberOfBytesRead) __out_data_source(FILE) BYTE* pbyInputReport, _In_ DWORD nNumberOfBytesToRead, _Out_opt_ LPDWORD lpNumberOfBytesRead = nullptr) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_h != INVALID_HANDLE_VALUE);

    return ReadFile(*this, pbyInputReport, nNumberOfBytesToRead, lpNumberOfBytesRead, nullptr) != FALSE;
  }

  _Success_(return != false) bool OutIn(_In_ bool bWrite, _In_ bool bRead) noexcept
  {
    //Reset the receive buffer
    m_RxBuffer = {};

#pragma warning(suppress: 26446)
    ATLASSERT(m_TxBuffer[0] == 0); //Should always be 0

    if (bWrite)
    {
#pragma warning(suppress: 26472)
      if (!WriteOutputReport(m_TxBuffer.data(), static_cast<DWORD>(m_TxBuffer.size())))
        return false;
    }
    if (bRead)
    {
      DWORD dwBytesRead = 0;
#pragma warning(suppress: 26472)
      if (!ReadInputReport(m_RxBuffer.data(), static_cast<DWORD>(m_RxBuffer.size()), &dwBytesRead))
        return false;
      if (dwBytesRead != m_RxBuffer.size())
      {
        SetLastError(ERROR_WRITE_FAULT);
        return false;
      }
    }
    return true;
  }

//Member variables
  std::array<BYTE, 9> m_RxBuffer{};
  std::array<BYTE, 9> m_TxBuffer{};
  BYTE m_LastDigitalOutValue{};
  int m_nCardAddress{-1};
  bool m_bNewMode{false};
};


//Client wrapper for the K8055D DLL using LoadLibrary / GetProcAddress
class C8055NWRAPPERS_EXT_CLASS C8055NClientDevice
{
public:
//Constructors / Destructors
  C8055NClientDevice(LPCTSTR pszPath8055D = _T("K8055D.DLL")) noexcept
  {
    //Load the DLL and get the function pointers
    m_hK8055N = LoadLibrary(pszPath8055D);
    if (m_hK8055N == nullptr)
      return;
#pragma warning(suppress: 26473 26490)
    m_lpfnSearchDevices = reinterpret_cast<LPSEARCHDEVICES>(GetProcAddress(m_hK8055N, "SearchDevices"));
#pragma warning(suppress: 26490)
    m_lpfnSetCurrentDevice = reinterpret_cast<LPSETCURRENTDEVICE>(GetProcAddress(m_hK8055N, "SetCurrentDevice"));
#pragma warning(suppress: 26490)
    m_lpfnOpenDevice = reinterpret_cast<LPOPENDEVICE>(GetProcAddress(m_hK8055N, "OpenDevice"));
#pragma warning(suppress: 26490)
    m_lpfnCloseDevice = reinterpret_cast<LPCLOSEDEVICE>(GetProcAddress(m_hK8055N, "CloseDevice"));
#pragma warning(suppress: 26490)
    m_lpfnReadAnalogChannel = reinterpret_cast<LPREADANALOGCHANNEL>(GetProcAddress(m_hK8055N, "ReadAnalogChannel"));
#pragma warning(suppress: 26490)
    m_lpfnReadAllAnalog = reinterpret_cast<LPREADALLANALOG>(GetProcAddress(m_hK8055N, "ReadAllAnalog"));
#pragma warning(suppress: 26490)
    m_lpfnOutputAnalogChannel = reinterpret_cast<LPOUTPUTANALOGCHANNEL>(GetProcAddress(m_hK8055N, "OutputAnalogChannel"));
#pragma warning(suppress: 26490)
    m_lpfnOutputAllAnalog = reinterpret_cast<LPOUTPUTALLANALOG>(GetProcAddress(m_hK8055N, "OutputAllAnalog"));
#pragma warning(suppress: 26490)
    m_lpfnClearAnalogChannel = reinterpret_cast<LPCLEARANALOGCHANNEL>(GetProcAddress(m_hK8055N, "ClearAnalogChannel"));
#pragma warning(suppress: 26490)
    m_lpfnClearAllAnalog = reinterpret_cast<LPCLEARALLANALOG>(GetProcAddress(m_hK8055N, "ClearAllAnalog"));
#pragma warning(suppress: 26490)
    m_lpfnSetAnalogChannel = reinterpret_cast<LPSETANALOGCHANNEL>(GetProcAddress(m_hK8055N, "SetAnalogChannel"));
#pragma warning(suppress: 26490)
    m_lpfnSetAllAnalog = reinterpret_cast<LPSETALLANALOG>(GetProcAddress(m_hK8055N, "SetAllAnalog"));
#pragma warning(suppress: 26490)
    m_lpfnWriteAllDigital = reinterpret_cast<LPWRITEALLDIGITAL>(GetProcAddress(m_hK8055N, "WriteAllDigital"));
#pragma warning(suppress: 26490)
    m_lpfnClearDigitalChannel = reinterpret_cast<LPCLEARDIGITALCHANNEL>(GetProcAddress(m_hK8055N, "ClearDigitalChannel"));
#pragma warning(suppress: 26490)
    m_lpfnClearAllDigital = reinterpret_cast<LPCLEARALLDIGITAL>(GetProcAddress(m_hK8055N, "ClearAllDigital"));
#pragma warning(suppress: 26490)
    m_lpfnSetDigitalChannel = reinterpret_cast<LPSETDIGITALCHANNEL>(GetProcAddress(m_hK8055N, "SetDigitalChannel"));
#pragma warning(suppress: 26490)
    m_lpfnSetAllDigital = reinterpret_cast<LPSETALLDIGITAL>(GetProcAddress(m_hK8055N, "SetAllDigital"));
#pragma warning(suppress: 26490)
    m_lpfnReadDigitalChannel = reinterpret_cast<LPREADDIGITALCHANNEL>(GetProcAddress(m_hK8055N, "ReadDigitalChannel"));
#pragma warning(suppress: 26473 26490)
    m_lpfnReadAllDigital = reinterpret_cast<LPREADALLDIGITAL>(GetProcAddress(m_hK8055N, "ReadAllDigital"));
#pragma warning(suppress: 26490)
    m_lpfnReadCounter = reinterpret_cast<LPREADCOUNTER>(GetProcAddress(m_hK8055N, "ReadCounter"));
#pragma warning(suppress: 26490)
    m_lpfnReadCounter16 = reinterpret_cast<LPREADCOUNTER16>(GetProcAddress(m_hK8055N, "ReadCounter16"));
#pragma warning(suppress: 26490)
    m_lpfnReadAll = reinterpret_cast<LPREADALL>(GetProcAddress(m_hK8055N, "ReadAll"));
#pragma warning(suppress: 26490)
    m_lpfnResetCounter = reinterpret_cast<LPRESETCOUNTER>(GetProcAddress(m_hK8055N, "ResetCounter"));
#pragma warning(suppress: 26490)
    m_lpfnSetCounterDebounceTime = reinterpret_cast<LPSETCOUNTERDEBOUNCETIME>(GetProcAddress(m_hK8055N, "SetCounterDebounceTime"));
#pragma warning(suppress: 26490)
    m_lpfnConnected = reinterpret_cast<LPCONNECTED>(GetProcAddress(m_hK8055N, "Connected"));
#pragma warning(suppress: 26490)
    m_lpfnSetPWM = reinterpret_cast<LPSETPWM>(GetProcAddress(m_hK8055N, "SetPWM"));
#pragma warning(suppress: 26490)
    m_lpfnReadBackAnalogOut = reinterpret_cast<LPREADBACKANALOGOUT>(GetProcAddress(m_hK8055N, "ReadBackAnalogOut"));
#pragma warning(suppress: 26473 26490)
    m_lpfnReadBackDigitalOut = reinterpret_cast<LPREADBACKDIGITALOUT>(GetProcAddress(m_hK8055N, "ReadBackDigitalOut"));
#pragma warning(suppress: 26473 26490)
    m_lpfnVersion = reinterpret_cast<LPVERSION>(GetProcAddress(m_hK8055N, "Version"));
#pragma warning(suppress: 26473 26490)
    m_lpfnCardVersion = reinterpret_cast<LPCARDVERSION>(GetProcAddress(m_hK8055N, "CardVersion"));
  }
  C8055NClientDevice(_In_ const C8055NClientDevice& device) = delete;
  C8055NClientDevice(_In_ C8055NClientDevice&& device) = delete;
  ~C8055NClientDevice()
  {
    //Unload the DLL if required
    if (m_hK8055N != nullptr)
    {
      FreeLibrary(m_hK8055N);
      m_hK8055N = nullptr;
    }
  }

//Methods
  C8055NClientDevice& operator=(_In_ const C8055NClientDevice& device) = delete;
  C8055NClientDevice& operator=(_In_ C8055NClientDevice&& device) = delete;
  int SearchDevices() noexcept
  {
    if (m_lpfnSearchDevices == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return -1;
    }
    return m_lpfnSearchDevices();
  }
  int SetCurrentDevice(_In_ int nCardNumber) noexcept
  {
    if (m_lpfnSetCurrentDevice == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return -1;
    }
    return m_lpfnSetCurrentDevice(nCardNumber);
  }
  int OpenDevice(_In_ int nCardNumber) noexcept
  {
    if (m_lpfnOpenDevice == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return -1;
    }
    return m_lpfnOpenDevice(nCardNumber);
  }
  void CloseDevice() noexcept
  {
    if (m_lpfnCloseDevice == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    m_lpfnCloseDevice();
  }
  int ReadAnalogChannel(_In_ int nChannel) noexcept
  {
    if (m_lpfnCloseDevice == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return 0;
    }
    return m_lpfnReadAnalogChannel(nChannel);
  }
  void ReadAllAnalog(_Inout_ int* pData1, _Inout_ int* pData2) noexcept
  {
    if (m_lpfnReadAllAnalog == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    m_lpfnReadAllAnalog(pData1, pData2);
  }
  void OutputAnalogChannel(_In_ int nChannel, _In_ int Data) noexcept
  {
    if (m_lpfnReadAllAnalog == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    m_lpfnOutputAnalogChannel(nChannel, Data);
  }
  void OutputAllAnalog(_In_ int Data1, _In_ int Data2) noexcept
  {
    if (m_lpfnOutputAllAnalog == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    m_lpfnOutputAllAnalog(Data1, Data2);
  }
  void ClearAnalogChannel(_In_ int nChannel) noexcept
  {
    if (m_lpfnClearAnalogChannel == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    m_lpfnClearAnalogChannel(nChannel);
  }
  void ClearAllAnalog() noexcept
  {
    if (m_lpfnClearAllAnalog == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    m_lpfnClearAllAnalog();
  }
  void SetAnalogChannel(_In_ int nChannel) noexcept
  {
    if (m_lpfnSetAnalogChannel == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    m_lpfnSetAnalogChannel(nChannel);
  }
  void SetAllAnalog() noexcept
  {
    if (m_lpfnSetAllAnalog == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    m_lpfnSetAllAnalog();
  }
  void WriteAllDigital(_In_ int Data) noexcept
  {
    if (m_lpfnWriteAllDigital == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    m_lpfnWriteAllDigital(Data);
  }
  void ClearDigitalChannel(_In_ int nChannel) noexcept
  {
    if (m_lpfnClearDigitalChannel == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    m_lpfnClearDigitalChannel(nChannel);
  }
  void ClearAllDigital() noexcept
  {
    if (m_lpfnClearAllDigital == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    m_lpfnClearAllDigital();
  }
  void SetDigitalChannel(_In_ int nChannel) noexcept
  {
    if (m_lpfnSetDigitalChannel == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    m_lpfnSetDigitalChannel(nChannel);
  }
  void SetAllDigital() noexcept
  {
    if (m_lpfnSetAllDigital == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    m_lpfnSetAllDigital();
  }
  BOOL ReadDigitalChannel(_In_ int nChannel) noexcept
  {
    if (m_lpfnReadDigitalChannel == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return FALSE;
    }
    return m_lpfnReadDigitalChannel(nChannel);
  }
  int ReadAllDigital() noexcept
  {
    if (m_lpfnReadAllDigital == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return 0;
    }
    return m_lpfnReadAllDigital();
  }
  int ReadCounter(_In_ int nCounter) noexcept
  {
    if (m_lpfnReadCounter == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return 0;
    }
    return m_lpfnReadCounter(nCounter);
  }
  int ReadCounter16(_In_ int nCounter) noexcept
  {
    if (m_lpfnReadCounter16 == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return 0;
    }
    return m_lpfnReadCounter16(nCounter);
  }
  void ReadAll(_Inout_updates_(8) BYTE* pData) noexcept
  {
    if (m_lpfnReadAll == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    return m_lpfnReadAll(pData);
  }
  void ResetCounter(_In_ int nCounter) noexcept
  {
    if (m_lpfnResetCounter == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    return m_lpfnResetCounter(nCounter);
  }
  void SetCounterDebounceTime(_In_ int nCounter, _In_ int nDebounceTime) noexcept
  {
    if (m_lpfnSetCounterDebounceTime == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    return m_lpfnSetCounterDebounceTime(nCounter, nDebounceTime);
  }
  _Success_(return != FALSE) BOOL Connected() noexcept
  {
    if (m_lpfnConnected == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return 0;
    }
    return m_lpfnConnected();
  }
  void SetPWM(_In_ int nChannel, _In_ int nData, _In_ int nFrequency) noexcept
  {
    if (m_lpfnSetPWM == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    m_lpfnSetPWM(nChannel, nData, nFrequency);
  }
  void ReadBackAnalogOut(_Inout_updates_(2) int* pBuffer) noexcept
  {
    if (m_lpfnReadBackAnalogOut == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return;
    }
    m_lpfnReadBackAnalogOut(pBuffer);
  }
  int ReadBackDigitalOut() noexcept
  {
    if (m_lpfnReadBackDigitalOut == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return 0;
    }
    return m_lpfnReadBackDigitalOut();
  }
  int Version() noexcept
  {
    if (m_lpfnVersion == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return 0;
    }
    return m_lpfnVersion();
  }
  int CardVersion() noexcept
  {
    if (m_lpfnCardVersion == nullptr)
    {
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return 0;
    }
    return m_lpfnCardVersion();
  }

protected:
//Typedefs of the function pointers
  using LPSEARCHDEVICES = int (WINAPI *)();
  using LPSETCURRENTDEVICE = int (WINAPI *)(int);
  using LPOPENDEVICE = int (WINAPI *)(int);
  using LPCLOSEDEVICE = void (WINAPI *)();
  using LPREADANALOGCHANNEL = int (WINAPI *)(int);
  using LPREADALLANALOG = void (WINAPI *)(int*, int*);
  using LPOUTPUTANALOGCHANNEL = void (WINAPI *)(int, int);
  using LPOUTPUTALLANALOG = void (WINAPI *)(int, int);
  using LPCLEARANALOGCHANNEL = void (WINAPI *)(int);
  using LPCLEARALLANALOG = void (WINAPI *)();
  using LPSETANALOGCHANNEL = void (WINAPI *)(int);
  using LPSETALLANALOG = void (WINAPI *)();
  using LPWRITEALLDIGITAL = void (WINAPI *)(int);
  using LPCLEARDIGITALCHANNEL = void (WINAPI *)(int);
  using LPCLEARALLDIGITAL = void (WINAPI *)();
  using LPSETDIGITALCHANNEL = void (WINAPI *)(int);
  using LPSETALLDIGITAL = void (WINAPI *)();
  using LPREADDIGITALCHANNEL = BOOL (WINAPI *)(int);
  using LPREADALLDIGITAL = int (WINAPI *)();
  using LPREADCOUNTER = int (WINAPI *)(int);
  using LPREADCOUNTER16 = int (WINAPI *)(int);
  using LPREADALL = void (WINAPI *)(BYTE*);
  using LPRESETCOUNTER = void (WINAPI *)(int);
  using LPSETCOUNTERDEBOUNCETIME = void (WINAPI *)(int, int);
  using LPCONNECTED = BOOL (WINAPI *)();
  using LPSETPWM = void (WINAPI*)(int, int, int);
  using LPREADBACKANALOGOUT = void (WINAPI *)(int*);
  using LPREADBACKDIGITALOUT = int (WINAPI *)();
  using LPVERSION = int (WINAPI *)();
  using LPCARDVERSION = int (WINAPI *)();

//Member variables
  HINSTANCE m_hK8055N{nullptr};
  LPSEARCHDEVICES m_lpfnSearchDevices{nullptr};
  LPSETCURRENTDEVICE m_lpfnSetCurrentDevice{nullptr};
  LPOPENDEVICE m_lpfnOpenDevice{nullptr};
  LPCLOSEDEVICE m_lpfnCloseDevice{nullptr};
  LPREADANALOGCHANNEL m_lpfnReadAnalogChannel{nullptr};
  LPREADALLANALOG m_lpfnReadAllAnalog{nullptr};
  LPOUTPUTANALOGCHANNEL m_lpfnOutputAnalogChannel{nullptr};
  LPOUTPUTALLANALOG m_lpfnOutputAllAnalog{nullptr};
  LPCLEARANALOGCHANNEL m_lpfnClearAnalogChannel{nullptr};
  LPCLEARALLANALOG m_lpfnClearAllAnalog{nullptr};
  LPSETANALOGCHANNEL m_lpfnSetAnalogChannel{nullptr};
  LPSETALLANALOG m_lpfnSetAllAnalog{nullptr};
  LPWRITEALLDIGITAL m_lpfnWriteAllDigital{nullptr};
  LPCLEARDIGITALCHANNEL m_lpfnClearDigitalChannel{nullptr};
  LPCLEARALLDIGITAL m_lpfnClearAllDigital{nullptr};
  LPSETDIGITALCHANNEL m_lpfnSetDigitalChannel{nullptr};
  LPSETALLDIGITAL m_lpfnSetAllDigital{nullptr};
  LPREADDIGITALCHANNEL m_lpfnReadDigitalChannel{nullptr};
  LPREADALLDIGITAL m_lpfnReadAllDigital{nullptr};
  LPREADCOUNTER m_lpfnReadCounter{nullptr};
  LPREADCOUNTER16 m_lpfnReadCounter16{ nullptr };
  LPREADALL m_lpfnReadAll{nullptr};
  LPRESETCOUNTER m_lpfnResetCounter{nullptr};
  LPSETCOUNTERDEBOUNCETIME m_lpfnSetCounterDebounceTime{nullptr};
  LPCONNECTED m_lpfnConnected{nullptr};
  LPSETPWM m_lpfnSetPWM{nullptr};
  LPREADBACKANALOGOUT m_lpfnReadBackAnalogOut{nullptr};
  LPREADBACKDIGITALOUT m_lpfnReadBackDigitalOut{nullptr};
  LPVERSION m_lpfnVersion{nullptr};
  LPCARDVERSION m_lpfnCardVersion{nullptr};
};


#define CHECK_DEVICE_IS_OPEN_RETURN_0()                                                              \
if (m_Device.operator HANDLE() == INVALID_HANDLE_VALUE)                                              \
{                                                                                                    \
  ATLASSERT(_T("You forgot to call OpenDevice or SetCurrentDevice before calling this function\n")); \
  SetLastError(ERROR_OPEN_FAILED);                                                                   \
  return 0;                                                                                          \
}                                                                                                    \

#define CHECK_DEVICE_IS_OPEN_RETURN_NOTHING()                                                        \
if (m_Device.operator HANDLE() == INVALID_HANDLE_VALUE)                                              \
{                                                                                                    \
  ATLASSERT(_T("You forgot to call OpenDevice or SetCurrentDevice before calling this function\n")); \
  SetLastError(ERROR_OPEN_FAILED);                                                                   \
  return;                                                                                            \
}                                                                                                    \


//Client wrapper which is similar to C8055NClientDevice but uses C8055NDevice class directly.
//This class is used in the implementation of the K8055N.DLL
class C8055NWRAPPERS_EXT_CLASS C8055NClientDevice2
{
public:
  //Constructors / Destructors
  C8055NClientDevice2() noexcept = default;
  C8055NClientDevice2(_In_ const C8055NClientDevice2& device) = delete;
  C8055NClientDevice2(_In_ C8055NClientDevice2&& device) = delete;
  ~C8055NClientDevice2() = default;

  //Methods
  C8055NClientDevice2& operator=(_In_ const C8055NClientDevice2& device) = delete;
  C8055NClientDevice2& operator=(_In_ C8055NClientDevice2&& device) = delete;

  int SearchDevices()
  {
    const bool bSuccess = HID::C8055NDevice::Enumerate(m_FoundDevices);
    if (!bSuccess)
      return 0; //No devices found
    int nReturn = 0;
    for (const auto& device : m_FoundDevices)
    {
      if (HID::CDevice::GetProductID(device.c_str()) == 0x5500)
        nReturn |= 1;
      if (HID::CDevice::GetProductID(device.c_str()) == 0x5501)
        nReturn |= 2;
      if (HID::CDevice::GetProductID(device.c_str()) == 0x5502)
        nReturn |= 4;
      if (HID::CDevice::GetProductID(device.c_str()) == 0x5503)
        nReturn |= 8;
    }
    return nReturn;
  }

  int SetCurrentDevice(_In_ int nCardNumber)
  {
    //Find the correct item in the m_FoundDevices array
    const auto iter = std::find_if(m_FoundDevices.cbegin(), m_FoundDevices.cend(), [nCardNumber](const auto& element) noexcept
      {
        return (HID::CDevice::GetProductID(element.c_str()) - 0x5500) == nCardNumber;
      });
    if (iter == m_FoundDevices.cend())
      return -1;

    //Open up the device
    m_Device.Close();
    if (!m_Device.Open(iter->c_str()))
      return -1;

    //Return the indicator that the device opened successfully
    return nCardNumber;
  }

  int OpenDevice(_In_ int nCardNumber)
  {
    //Re-enumerate the devices
    if (!SearchDevices())
      return -1;

    //Delegate to SetCurrentDevice
    return SetCurrentDevice(nCardNumber);
  }

  void CloseDevice() noexcept
  {
    m_Device.Close();
    m_FoundDevices.clear();
  }

  int ReadAnalogChannel(_In_ int nChannel) noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_0()

    BYTE byValue{0};
    if (!m_Device.ReadAnalogChannel(nChannel, byValue))
      return 0;
    return byValue;
  }

  void ReadAllAnalog(_Inout_ int* pData1, _Inout_ int* pData2) noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_NOTHING();
    if ((pData1 == nullptr) || (pData2 == nullptr))
    {
      ATLASSERT(_T("Both parameters to this function should be non null\n"));
      SetLastError(ERROR_INVALID_PARAMETER);
      return;
    }

    BYTE byData1{0};
    BYTE byData2{0};
    if (!m_Device.ReadAllAnalog(byData1, byData2))
      return;
    *pData1 = byData1;
    *pData2 = byData2;
  }

  void OutputAnalogChannel(_In_ int nChannel, _In_ int nData) noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_NOTHING();

#pragma warning(suppress: 26472)
    m_Device.OutputAnalogChannel(nChannel, static_cast<BYTE>(nData));
  }

  void OutputAllAnalog(_In_ int Data1, _In_ int Data2) noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_NOTHING();

#pragma warning(suppress: 26472)
    m_Device.OutputAllAnalog(static_cast<BYTE>(Data1), static_cast<BYTE>(Data2));
  }

  void ClearAnalogChannel(_In_ int nChannel) noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_NOTHING();

    m_Device.ClearAnalogChannel(nChannel);
  }

  void ClearAllAnalog() noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_NOTHING();

    m_Device.ClearAllAnalog();
  }

  void SetAnalogChannel(_In_ int nChannel) noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_NOTHING();

    m_Device.SetAnalogChannel(nChannel);
  }

  void SetAllAnalog() noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_NOTHING();

    m_Device.SetAllAnalog();
  }

  void WriteAllDigital(_In_ int Data) noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_NOTHING();

#pragma warning(suppress: 26472)
    m_Device.WriteAllDigital(static_cast<BYTE>(Data));
  }

  void ClearDigitalChannel(_In_ int nChannel) noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_NOTHING();

    m_Device.ClearDigitalChannel(nChannel);
  }

  void ClearAllDigital() noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_NOTHING();

    m_Device.ClearAllDigital();
  }

  void SetDigitalChannel(_In_ int nChannel) noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_NOTHING();

    m_Device.SetDigitalChannel(nChannel);
  }

  void SetAllDigital() noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_NOTHING();

    m_Device.SetAllDigital();
  }

  _Success_(return != FALSE) BOOL ReadDigitalChannel(_In_ int nChannel) noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_0();

    bool bValue{false};
    if (!m_Device.ReadDigitalChannel(nChannel, bValue))
      return FALSE;
    return bValue != FALSE;
  }

  int ReadAllDigital() noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_0();

    int nValue{0};
    if (!m_Device.ReadAllDigital(nValue))
      return 0;
    return nValue;
  }

  int ReadCounter(_In_ int nCounter) noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_0();

    DWORD dwValue{0};
    if (!m_Device.ReadCounter(nCounter, dwValue))
      return 0;
    return dwValue;
  }

  int ReadCounter16(_In_ int nCounter) noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_0();

    USHORT nValue{0};
    if (!m_Device.ReadCounter16(nCounter, nValue))
      return 0;
    return nValue;
  }

  void ReadAll(_Inout_updates_(8) BYTE* pData) noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_NOTHING();

    m_Device.ReadAll(pData);
  }

  void ResetCounter(_In_ int nCounter) noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_NOTHING();

    m_Device.ResetCounter(nCounter);
  }

  void SetCounterDebounceTime(_In_ int nCounter, _In_ int nDebounceTime) noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_NOTHING();

    m_Device.SetCounterDebounceTime(nCounter, nDebounceTime);
  }

  _Success_(return != FALSE) BOOL Connected() noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_0();

    return m_Device.Connected() != FALSE;
  }

  void SetPWM(_In_ int nChannel, _In_ int nData, _In_ int nFrequency) noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_NOTHING();

#pragma warning(suppress: 26472)
    m_Device.SetPWM(nChannel, static_cast<BYTE>(nData), static_cast<BYTE>(nFrequency));
  }

  void ReadBackAnalogOut(_Inout_updates_(2) int* pBuffer) noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_NOTHING();
    if (pBuffer == nullptr)
    {
      SetLastError(ERROR_INVALID_PARAMETER);
      return;
    }

    int nValue1{0};
    int nValue2{0};
    if (!m_Device.ReadBackAnalogOut(nValue1, nValue2))
      return;
#pragma warning(suppress: 26481)
    pBuffer[0] = nValue1;
#pragma warning(suppress: 26481)
    pBuffer[1] = nValue2;
  }

  int ReadBackDigitalOut() noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_0();

    int nValue{0};
    if (!m_Device.ReadBackDigitalOut(nValue))
      return 0;
    return nValue;
  }

  int CardVersion() noexcept
  {
    //Validate our parameters
    CHECK_DEVICE_IS_OPEN_RETURN_0();

    C8055NDevice::CARD_VERSION cardVersion{HID::C8055NDevice::CARD_VERSION::UNKNOWN};
    if (!m_Device.CardVersion(cardVersion))
      return static_cast<int>(HID::C8055NDevice::CARD_VERSION::UNKNOWN);
    return static_cast<int>(cardVersion);
  }

protected:
//Member variables
  StringArray m_FoundDevices;
  C8055NDevice m_Device;
};


}; //namespace HID

#endif //#ifndef __8055NWRAPPERS_H__
